.collapse-card-container {
    padding: 10px;
    margin: 0 0 10px 0;
    border: 1px solid #d2d2d2;
    border-radius: 6px;
    display: flex;
    justify-content: space-between;
}

.collapse-card-data {
    display: flex;
    flex-direction: column;
    width: 100%;
}

.component-success {
    color: #146c43;
    background-color: #d1e7dd;
}

.component-failure {
    color: #b02a37;
    background-color: #f8d7da;
}

.reports-container {
    display: flex;
}

.btn-type-container {
    display: flex;
    flex-direction: column;
    padding: 0px 15px;
}

.report-title {
    border-bottom: 1px solid #d2d2d2;
    padding-bottom: 10px;
}

.down-arrow {
    transform: translate(4px, -4px);
}

.expand-details-btn {
    display: flex;
    width: 138px;
    padding: 5px !important;
    margin-top: 3px;
}

.terminal {
    background-color: black;
    color: white;
    font-family: monospace;
    font-size: 14px;
    width: 100%;
    border-radius: 10px;
    overflow: hidden;
    position: relative;
    margin-top: 10px;
}

.terminal-header {
    background-color: #444;
    height: 20px;
    padding: 5px;
    display: flex;
    align-items: center;
    justify-content: center;
}

.terminal-title {
    margin-left: 5px;
}

.terminal-body {
    padding: 10px;
    height: 100%;
    overflow-y: scroll;
}

.terminal-line {
    margin-bottom: 5px;
}

.terminal-text {
    color: #4ec9b0;
}

.component-indent {
    margin-left: 10px;
}

.btn-success {
    --bs-btn-color: #146c43 !important;
    --bs-btn-bg: #d1e7dd !important;
    --bs-btn-border-color: #d1e7dd !important;
    --bs-btn-hover-color: #146c43 !important;
    --bs-btn-hover-bg: #bed6cb !important;
    --bs-btn-hover-border-color: #146c43 !important;
    --bs-btn-focus-shadow-rgb: 60, 153, 110 !important;
    --bs-btn-active-color: #fff !important;
    --bs-btn-active-bg: #146c43 !important;
    --bs-btn-active-border-color: #13653f !important;
    --bs-btn-active-shadow: inset 0 3px 5px rgba(0, 0, 0, 0.125) !important;
    --bs-btn-disabled-color: #fff !important;
    --bs-btn-disabled-bg: #d1e7dd !important;
    --bs-btn-disabled-border-color: #d1e7dd !important;
}

.btn-danger {
    --bs-btn-color: #b02a37 !important;
    --bs-btn-bg: #f8d7da !important;
    --bs-btn-border-color: #f8d7da !important;
    --bs-btn-hover-color: #b02a37 !important;
    --bs-btn-hover-bg: #e3b2b7 !important;
    --bs-btn-hover-border-color: #b02a37 !important;
    --bs-btn-focus-shadow-rgb: 60, 153, 110 !important;
    --bs-btn-active-color: #fff !important;
    --bs-btn-active-bg: #b02a37 !important;
    --bs-btn-active-border-color: #b02a37 !important;
    --bs-btn-active-shadow: inset 0 3px 5px rgba(0, 0, 0, 0.125) !important;
    --bs-btn-disabled-color: #fff !important;
    --bs-btn-disabled-bg: #f8d7da !important;
    --bs-btn-disabled-border-color: #f8d7da !important;
}

.btn-warning {
    --bs-btn-color: #5a5a2d !important;
    --bs-btn-bg: #f3f390 !important;
    --bs-btn-border-color: #f3f390 !important;
    --bs-btn-hover-color: #5a5a2d !important;
    --bs-btn-hover-bg: #d4d470 !important;
    --bs-btn-hover-border-color: #5a5a2d !important;
    --bs-btn-focus-shadow-rgb: 60, 153, 110 !important;
    --bs-btn-active-color: #fff !important;
    --bs-btn-active-bg: #5a5a2d !important;
    --bs-btn-active-border-color: #f5ba0c !important;
    --bs-btn-active-shadow: inset 0 3px 5px rgba(0, 0, 0, 0.125) !important;
    --bs-btn-disabled-color: #fff !important;
    --bs-btn-disabled-bg: #f3f390 !important;
    --bs-btn-disabled-border-color: #f3f390 !important;
}
